/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

import at.jku.fim.rubanetra.config.model.PortSpecification;
import at.jku.fim.rubanetra.output.OutputWriterStrategy;
import at.jku.fim.rubanetra.protocol.activity.Activity;
import at.jku.fim.rubanetra.protocol.activity.dns.DnsActivity;
import at.jku.fim.rubanetra.protocol.mapper.KrakenTransportLayerMappingFactory;
import org.junit.Test;
import org.xbill.DNS.*;
import org.xml.sax.SAXException;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.stream.XMLStreamException;
import javax.xml.transform.TransformerException;
import javax.xml.xpath.XPathExpressionException;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Collection;

import static org.junit.Assert.*;

/**
 * Created by stefan on 1/14/14.
 */
public class DnsTest extends TestBase {
    private static final String pcapDnssecTcp = "./captures/04.dnssec.over.tcp.pcapng";
    private final String pcapDns = "./captures/03.skype.irc.cap";
    private final String pcapDnsTcp = "captures/05.dns.over.tcp.cap";
    private final String pcapDns40min = "captures/10.dns.40min.pcapng";

    public void initializeDefaultDnsSetting(String pcapPath) {
        initializePcapHandler(pcapPath);
        addDefaultDnsSupport();
        setRubanetraSystemConfiguration(sysBuilder.build());
    }

    @Test
    public void initializeDnsConfigTest() {
        initializeDefaultDnsSetting(pcapDns);
        assertNotNull(super.pcapHandler);
        assertTrue(pcapHandler.readNextPcapEntry(null));
        while (pcapHandler.readNextPcapEntry(null)) ;
        assertFalse(pcapHandler.readNextPcapEntry(null));
    }

    @Test
    public void dnsActivityOutputTest() throws IOException, ParserConfigurationException, SAXException, XPathExpressionException, TransformerException, XMLStreamException {
        File tempFile = File.createTempFile("dnsactivityoutputtest", ".json");
        assertTrue(tempFile.exists() && tempFile.canWrite());
        tempFile.deleteOnExit();
        log.info("Created temp file: {}", tempFile.getAbsolutePath());
        sysBuilder.setOutputFormat("plaso").setOutputStream(new FileOutputStream(tempFile));
//        sysBuilder.setOutputFormat("xml").setOutputStream(System.out);

        initializeDefaultDnsSetting(pcapDnsTcp);

        while (pcapHandler.readNextPcapEntry(droolsHandler)) ;
        final Collection<?> dnsActivities = systemConfiguration.getKieSession().getObjects(object -> object instanceof DnsActivity);
        assertFalse(dnsActivities.isEmpty());
        int dnsActivitiesSize = dnsActivities.size();
        droolsHandler.close();
        assertTrue(tempFile.canRead());
        assertTrue(tempFile.length() > 1);

        // TODO: Parse XML output and perform validation
    }

    @Test
    public void dnsActivityExpirationOutputTest() throws IOException {
        sysBuilder.setOutputFormat("callback").setOutputStream(System.out).setOutputWriterStrategy(new OutputWriterStrategy() {
            int nrOfDnsActivities = 0;

            @Override
            public void writeActivity(Activity activity) {
                if (activity instanceof DnsActivity) {
                    // currently, certain DHCP packets are considered to be empty DNS messages by dnsjava
                    // as soon as this is fixed, 184 DNS-activities should be encountered.
                    // frame numbers: 49,50,235.236
                    long frameNumber = ((DnsActivity) activity).getPcapActivity().getFrameNumber();
                    if (frameNumber != 49 && frameNumber != 50 && frameNumber != 235 && frameNumber != 236) {
                        nrOfDnsActivities++;
                    }
                }
            }

            @Override
            public void closeWriter() throws IOException {
                assertEquals(184, nrOfDnsActivities);
            }
        });
        initializeDefaultDnsSetting(pcapDns40min);

        while (pcapHandler.readNextPcapEntry(droolsHandler)) ;
        final Collection<?> dnsActivities = systemConfiguration.getKieSession().getObjects(object -> object instanceof DnsActivity);
        droolsHandler.close();
    }

    @Test
    public void dnsActivitymultipleBindings() throws IOException {
        sysBuilder.setOutputFormat("callback").setOutputStream(System.out).setOutputWriterStrategy(new OutputWriterStrategy() {
            int nrOfDnsActivities = 0;

            @Override
            public void writeActivity(Activity activity) {
                if (activity instanceof DnsActivity) {
                    nrOfDnsActivities++;
                }
            }

            @Override
            public void closeWriter() throws IOException {
                assertEquals(6, nrOfDnsActivities);
            }
        });
        initializePcapHandler("captures/11.dns.ipv6.pcapng");


        protoBuilder.resetCurrentProtocolConfiguration();
        protoBuilder.bindProtocol("ethernet", "ipv4").bindProtocol("ethernet", "ipv6").bindProtocol("ipv4", "tcp")
                .bindProtocol("ipv4", "udp").bindProtocol("ipv6", "tcp").bindProtocol("ipv6", "udp")
                .bindProtocol("tcp", "dns").bindProtocol("udp", "dns");
        protoBuilder.setPcapHandler(pcapHandler);
        protoBuilder.setPortSpecification(PortSpecification.UniversalPortSpecification.ACCEPT_ALL);
        protoBuilder.setTransportProtocolMappingStrategy(KrakenTransportLayerMappingFactory.DESTINATION_PORT_STRATEGY);
        protoBuilder.setTransportProtocolMappingFactory(transportProtocolMappingFactory);
        sysBuilder.addProtocolSetting(protoBuilder.buildProtocolConfiguration());


        setRubanetraSystemConfiguration(sysBuilder.build());

        while (pcapHandler.readNextPcapEntry(droolsHandler)) ;
        final Collection<?> dnsActivities = systemConfiguration.getKieSession().getObjects(object -> object instanceof DnsActivity);
        droolsHandler.close();
    }

    @Test
    public void existsDnsActivityTest() {
        initializeDefaultDnsSetting(pcapDnsTcp);
        while (pcapHandler.readNextPcapEntry(droolsHandler)) ;
        final Collection<?> dnsActivities = systemConfiguration.getKieSession().getObjects(object -> object instanceof DnsActivity);
        assertFalse(dnsActivities.isEmpty());

    }

    @Test
    public void existsDnssecOverTcpTest() {
        initializeDefaultDnsSetting(pcapDnssecTcp);
        while (pcapHandler.readNextPcapEntry(droolsHandler)) ;
        final Collection<?> dnsActivities = systemConfiguration.getKieSession().getObjects(object -> object instanceof DnsActivity);

        assertTrue(!dnsActivities.isEmpty());

        int foundTcpRecord = 0;
        for (Object o : dnsActivities) {
            assertTrue(o instanceof DnsActivity);
            DnsActivity act = (DnsActivity) o;
            if (act.getPcapActivity().getFrameNumber() == 15) {
                assertNotNull(act.getDnsMessage());
                assertEquals(0x9a8c, act.getDnsMessageHeader().getID());
                assertNotNull(act.getDnsMessage().getQuestion());
                assertEquals(Name.fromConstantString("com."), act.getDnsMessage().getQuestion().getName());
                foundTcpRecord++;
            } else if (act.getPcapActivity().getFrameNumber() == 17) {
                assertNotNull(act.getDnsMessage());
                assertEquals(0x9a8c, act.getDnsMessageHeader().getID());
                assertNotNull(act.getDnsMessage().getQuestion());
                assertEquals(Name.fromConstantString("com."), act.getDnsMessage().getQuestion().getName());
                for (Record r : act.getAnswerRecords()) {
                    if (r.getType() == Type.DNSKEY) {
                        assertTrue(r instanceof DNSKEYRecord);
                        try {
                            assertNotNull(((DNSKEYRecord) r).getPublicKey());
                        } catch (DNSSEC.DNSSECException e) {
                            fail("Missing DNSSEC Public Key");
                        }
                    }
                }
                foundTcpRecord++;
            }
        }
        assertEquals(2, foundTcpRecord);
    }
}
